import hashlib
from pathlib import Path
from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render


def hash(path: Path, algo: str, *, _bufsize=2**18) -> str:
    # for python >= 3.11 we can use hashlib.file_digest
    if hasattr(hashlib, "file_digest"):
        with path.open("rb") as f:
            return hashlib.file_digest(f, algo, _bufsize=_bufsize).hexdigest()

    # otherwise we have to use our own backported implementation:
    h = hashlib.new(algo)
    buf = bytearray(_bufsize)
    view = memoryview(buf)
    with path.open("rb", buffering=0) as f:
        while size := f.readinto(buf):
            h.update(view[:size])
    return h.hexdigest()


def hash_first_chunk(path: Path, algo: str, *, chunksize=1024) -> str:
    h = hashlib.new(algo)
    with path.open("rb") as f:
        chunk = f.read(chunksize)
        h.update(chunk)
    return h.hexdigest()


@dataclass(config=ConfigDict(extra="forbid"))
class Hash:

    """Calculates the hash of a file.

    Attributes:
        algorithm (str): Any hashing algorithm available to python's `hashlib`.
            `md5` by default.

    Algorithms guaranteed to be available are
    `shake_256`, `sha3_256`, `sha1`, `sha3_224`, `sha384`, `sha512`, `blake2b`,
    `blake2s`, `sha256`, `sha224`, `shake_128`, `sha3_512`, `sha3_384` and `md5`.

    Depending on your python installation and installed libs there may be additional
    hash algorithms to chose from.

    To list the available algorithms on your installation run this in a python
    interpreter:

    ```py
    >>> import hashlib
    >>> hashlib.algorithms_available
    {'shake_256', 'whirlpool', 'mdc2', 'blake2s', 'sha224', 'shake_128', 'sha3_512',
    'sha3_224', 'sha384', 'md5', 'sha1', 'sha512_256', 'blake2b', 'sha256',
    'sha512_224', 'ripemd160', 'sha3_384', 'md4', 'sm3', 'sha3_256', 'md5-sha1',
    'sha512'}
    ```

    **Returns:**

    - `{hash}`:  The hash of the file.
    """

    algorithm: str = "md5"

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="hash",
        files=True,
        dirs=False,
    )

    def __post_init__(self):
        self._algorithm = Template.from_string(self.algorithm)

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None
        algo = render(self._algorithm, res.dict()).lower()
        result = hash(path=res.path, algo=algo)
        res.vars[self.filter_config.name] = result
        return True



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Set

from organize.utils import deep_merge

if TYPE_CHECKING:
    from .rule import Rule


@dataclass
class Resource:
    """
    A resource is created for each handled file (or folder) and then passed into the
    filters and actions pipeline.

    :param path:
        The path to the current file or folder
    :param basedir:
        The search location as given in rule->location->path
    :param rule:
        The rule which is currently executed
    :param rule_nr:
        The index of the rule in the config file
    :param vars:
        Filters and actions may add values to this dict which are then available for
        other filters and actions in the pipeline.
    :param walker_skip_files:
        Filters and actions may add pathes to this set which are then ignored for the
        rest of the rule.
    """

    path: Optional[Path]
    basedir: Optional[Path] = None
    rule: Optional[Rule] = None  # TODO: not optional?
    rule_nr: int = 0
    vars: Dict[str, Any] = field(default_factory=dict)
    walker_skip_pathes: Set[Path] = field(default_factory=set)

    def relative_path(self) -> Optional[Path]:
        if self.basedir is None:
            return self.path
        if self.path is None:
            return None
        return self.path.relative_to(self.basedir)

    def dict(self):
        return dict(
            path=self.path,
            basedir=self.basedir,
            location=self.basedir,
            relative_path=self.relative_path(),
            rule=self.rule.name if self.rule else None,
            rule_nr=self.rule_nr,
            **self.vars,
        )

    def deep_merge(self, key: str, data: Dict) -> None:
        """
        Convenience method for filters / actions to merge data into the `vars` dict.
        """
        prev = self.vars.get(key, dict())
        self.vars[key] = deep_merge(prev, data)

    # TODO: Caching for `is_file` and `is_dir`
    # TODO: provide a `from_direntry` constructor to speed things up
    def is_file(self) -> bool:
        if self.path is None:
            raise ValueError("No path given")
        return self.path.is_file()

    def is_dir(self) -> bool:
        if self.path is None:
            raise ValueError("No path given")
        return self.path.is_dir()

    def is_empty(self) -> bool:
        if self.path is None:
            raise ValueError("No path given")
        if self.is_file():
            return self.path.stat().st_size == 0
        elif self.is_dir():
            return not any(self.path.iterdir())
        raise ValueError("Unknown file type")



import subprocess
from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render

# TODO: Terminal waterfall: https://github.com/Textualize/rich/discussions/2985


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Shell:
    """
    Executes a shell command

    Attributes:
        cmd (str): The command to execute.
        run_in_simulation (bool):
            Whether to execute in simulation mode (default = false)
        ignore_errors (bool):
            Whether to continue on returncodes != 0.
        simulation_output (str):
            The value of `{shell.output}` if run in simulation
        simulation_returncode (int):
            The value of `{shell.returncode}` if run in simulation

    Returns

    - `{shell.output}` (`str`): The stdout of the executed process.
    - `{shell.returncode}` (`int`): The returncode of the executed process.
    """

    cmd: str
    run_in_simulation: bool = False
    ignore_errors: bool = False
    simulation_output: str = "** simulation **"
    simulation_returncode: int = 0

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="shell",
        standalone=True,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._cmd = Template.from_string(self.cmd)
        self._simulation_output = Template.from_string(self.simulation_output)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        full_cmd = render(self._cmd, res.dict())

        if not simulate or self.run_in_simulation:
            output.msg(res=res, msg=f"$ {full_cmd}", sender=self)
            try:
                call = subprocess.run(
                    full_cmd,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=True,
                )
            except subprocess.CalledProcessError as e:
                if not self.ignore_errors:
                    raise e

            res.vars[self.action_config.name] = {
                "output": call.stdout.decode("utf-8"),
                "returncode": call.returncode,
            }
        else:
            output.msg(
                res=res,
                msg=f"** not run in simulation ** $ {full_cmd}",
                sender=self,
            )
            res.vars[self.action_config.name] = {
                "output": render(self._simulation_output, res.dict()),
                "returncode": self.simulation_returncode,
            }



from __future__ import annotations

import shutil
from typing import TYPE_CHECKING, ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig

if TYPE_CHECKING:
    from pathlib import Path

    from organize.output import Output
    from organize.resource import Resource


def delete(path: Path):
    if path.is_dir():
        shutil.rmtree(path)
    else:
        path.unlink()


@dataclass(config=ConfigDict(extra="forbid"))
class Delete:

    """
    Delete a file from disk.

    Deleted files have no recovery option!
    Using the `Trash` action is strongly advised for most use-cases!
    """

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="delete",
        standalone=False,
        files=True,
        dirs=True,
    )

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        output.msg(res=res, msg=f"Deleting {res.path}", sender=self)
        if not simulate:
            delete(res.path)
        res.path = None



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, NamedTuple

from organize.output import Output
from organize.resource import Resource
from organize.template import render

if TYPE_CHECKING:
    from jinja2 import Template

# TODO: keep_newer, keep_older, keep_bigger, keep_smaller
ConflictMode = Literal["skip", "overwrite", "trash", "rename_new", "rename_existing"]


class ConflictResult(NamedTuple):
    skip_action: bool  # Whether to skip the current action
    use_dst: Path  # The Path to continue with


def next_free_name(dst: Path, template: Template) -> Path:
    """
    Increments {counter} in the template until the given resource does not exist.

    Attributes:
        dst (Path):
            The destination path.
        template (jinja2.Template):
            A jinja2 template with placeholders for {name}, {extension} and {counter}

    Raises:
        ValueError if no free name can be found with the given template.

    Returns:
        (Path) A path according to the given template that does not exist.
    """
    if not dst.exists():
        return dst
    counter = 2
    prev_candidate = None
    while True:
        args = dict(
            name=dst.stem,
            extension=dst.suffix,
            counter=counter,
        )
        new_name = render(template, args)
        candidate = dst.with_name(new_name)
        if not candidate.exists():
            return candidate
        if prev_candidate == candidate:
            raise ValueError(
                "Could not find a free filename for the given template. "
                'Maybe you forgot the "{counter}" placeholder?'
            )
        prev_candidate = candidate
        counter += 1


def resolve_conflict(
    dst: Path,
    res: Resource,
    conflict_mode: ConflictMode,
    rename_template: Template,
    simulate: bool,
    output: Output,
) -> ConflictResult:
    """
    Handle a conflict if `dst` already exists.
    """
    assert res.path is not None

    # no conflict, just continue with the action.
    if not dst.exists():
        return ConflictResult(skip_action=False, use_dst=dst)

    def _print(msg: str):
        output.msg(res=res, sender="conflict", msg=msg)

    _print(f'"{dst}" already exists! (Conflict mode is "{conflict_mode}")')

    if res.path.resolve() == dst.resolve():
        _print("Same resource: Skipped.")
        return ConflictResult(skip_action=True, use_dst=res.path)

    if conflict_mode == "trash":
        _print(f'Trash "{dst}"')
        if not simulate:
            from organize.actions.trash import trash

            trash(path=dst)
        return ConflictResult(skip_action=False, use_dst=dst)

    elif conflict_mode == "skip":
        _print("Skipped.")
        return ConflictResult(skip_action=True, use_dst=res.path)

    elif conflict_mode == "overwrite":
        _print(f"Overwriting {dst}.")
        if not simulate:
            from organize.actions.delete import delete

            delete(path=dst)
        return ConflictResult(skip_action=False, use_dst=dst)

    elif conflict_mode == "rename_new":
        new_path = next_free_name(
            dst=dst,
            template=rename_template,
        )
        return ConflictResult(skip_action=False, use_dst=new_path)

    elif conflict_mode == "rename_existing":
        new_path = next_free_name(
            dst=dst,
            template=rename_template,
        )
        _print('Renaming existing to: "{new_path.name}"')
        if not simulate:
            dst.rename(new_path)
        return ConflictResult(skip_action=False, use_dst=dst)

    raise ValueError("Unknown conflict_mode %s" % conflict_mode)



from typing import Any, Iterable, List, Mapping, TypeVar

from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated


def islist(x):
    return isinstance(x, Iterable) and not isinstance(x, (str, bytes, Mapping))


def _flatten(items):
    """Yield items from any nested iterable; see Reference."""
    for x in items:
        if islist(x):
            yield from _flatten(x)
        else:
            yield x


def flatten(x: Any):
    if x is None:
        return []
    if not islist(x):
        x = (x,)
    return list(_flatten(x))


T = TypeVar("T")
FlatList = Annotated[List[T], BeforeValidator(flatten)]



import os
from pathlib import Path
from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render

from .common.conflict import ConflictMode, resolve_conflict
from .common.target_path import prepare_target_path


def create_hardlink(target: Path, link: Path) -> None:
    # Path.hardlink_to needs Python >= 3.10
    os.link(src=target, dst=link)  # create a hardlink pointing to src named dst


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Hardlink:

    """Create a hardlink.

    Attributes:
        dest (str):
            The hardlink destination. If **dest** ends with a slash `/``, create the
            hardlink in the given directory. Can contain placeholders.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

        autodetect_folder (bool):
            In case you forget the ending slash "/" to indicate copying into a folder
            this settings will handle targets without a file extension as folders.
            If you really mean to copy to a file without file extension, set this to
            false.
            Default: true
    """

    dest: str
    on_conflict: ConflictMode = "rename_new"
    rename_template: str = "{name} {counter}{extension}"
    autodetect_folder: bool = True

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="hardlink",
        standalone=False,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._dest = Template.from_string(self.dest)
        self._rename_template = Template.from_string(self.rename_template)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        rendered = render(self._dest, res.dict())
        dst = prepare_target_path(
            src_name=res.path.name,
            dst=rendered,
            autodetect_folder=self.autodetect_folder,
            simulate=simulate,
        )

        skip_action, dst = resolve_conflict(
            dst=dst,
            res=res,
            conflict_mode=self.on_conflict,
            rename_template=self._rename_template,
            simulate=simulate,
            output=output,
        )
        if skip_action:
            return

        output.msg(res=res, msg=f"Creating hardlink at {dst}", sender=self)
        if not simulate:
            create_hardlink(target=res.path, link=dst)
        res.walker_skip_pathes.add(dst)



from typing import Any, ClassVar, List, Union

import simplematch
from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Name:
    """Match files and folders by name

    Attributes:
        match (str):
            A matching string in [simplematch-syntax](https://github.com/tfeldmann/simplematch)

        startswith (str):
            The filename must begin with the given string

        contains (str):
            The filename must contain the given string

        endswith (str):
            The filename (without extension) must end with the given string

        case_sensitive (bool):
            By default, the matching is case sensitive. Change this to False to use
            case insensitive matching.
    """

    match: str = "*"
    startswith: Union[str, List[str]] = ""
    contains: Union[str, List[str]] = ""
    endswith: Union[str, List[str]] = ""
    case_sensitive: bool = True

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="name",
        files=True,
        dirs=True,
    )

    def __post_init__(self, *args, **kwargs):
        self._matcher = simplematch.Matcher(
            self.match,
            case_sensitive=self.case_sensitive,
        )
        self.startswith = self.create_list(self.startswith, self.case_sensitive)
        self.contains = self.create_list(self.contains, self.case_sensitive)
        self.endswith = self.create_list(self.endswith, self.case_sensitive)

    def matches(self, name: str) -> bool:
        if not self.case_sensitive:
            name = name.lower()

        is_match = (
            self._matcher.test(name)
            and any(x in name for x in self.contains)
            and any(name.startswith(x) for x in self.startswith)
            and any(name.endswith(x) for x in self.endswith)
        )
        return is_match

    def pipeline(self, res: Resource, output: Output) -> bool:
        assert res.path is not None, "Does not support standalone mode"
        if res.is_dir():
            name = res.path.stem
        else:
            name, ext = res.path.stem, res.path.suffix
            if not name:
                name = ext
        result = self.matches(name)
        m = self._matcher.match(name)
        if not m:
            m = name

        res.vars[self.filter_config.name] = m
        return result

    @staticmethod
    def create_list(x: Union[int, str, List[Any]], case_sensitive: bool) -> List[str]:
        if isinstance(x, (int, float)):
            x = str(x)
        if isinstance(x, str):
            x = [x]
        x = [str(x) for x in x]
        if not case_sensitive:
            x = [x.lower() for x in x]
        return x



from typing import List, Set, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic.dataclasses import dataclass
from typing_extensions import Literal

from .validators import FlatList

DEFAULT_SYSTEM_EXCLUDE_FILES = {
    "thumbs.db",
    "desktop.ini",
    "~$*",
    ".DS_Store",
    ".localized",
}

DEFAULT_SYSTEM_EXCLUDE_DIRS = {
    ".git",
    ".svn",
}


@dataclass(config=ConfigDict(extra="forbid"))
class Location(BaseModel):
    path: FlatList[str]
    min_depth: int = 0
    max_depth: Union[Literal["inherit"], int, None] = "inherit"
    search: Literal["depth", "breadth"] = "breadth"
    exclude_files: Set[str] = Field(default_factory=set)
    exclude_dirs: Set[str] = Field(default_factory=set)
    system_exclude_files: Set[str] = Field(
        default_factory=lambda: DEFAULT_SYSTEM_EXCLUDE_FILES
    )
    system_exclude_dirs: Set[str] = Field(
        default_factory=lambda: DEFAULT_SYSTEM_EXCLUDE_DIRS
    )
    filter: Union[List[str], None] = None
    filter_dirs: Union[List[str], None] = None
    ignore_errors: bool = False

    @field_validator(
        "exclude_files",
        "exclude_dirs",
        "system_exclude_files",
        "system_exclude_dirs",
        mode="before",
    )
    @classmethod
    def ensure_set(cls, value):
        if isinstance(value, str):
            return set([value])
        return set(value)



from datetime import datetime, timezone
from pathlib import Path
from typing import ClassVar

from organize.filter import FilterConfig

from .common.timefilter import TimeFilter


def read_lastmodified(path: Path) -> datetime:
    return datetime.fromtimestamp(path.stat().st_mtime, tz=timezone.utc)


class LastModified(TimeFilter):

    """Matches files by last modified date

    Attributes:
        years (int): specify number of years
        months (int): specify number of months
        weeks (float): specify number of weeks
        days (float): specify number of days
        hours (float): specify number of hours
        minutes (float): specify number of minutes
        seconds (float): specify number of seconds
        mode (str):
            either 'older' or 'newer'. 'older' matches files / folders last modified before
            the given time, 'newer' matches files / folders last modified within the given
            time. (default = 'older')

    Returns:
        {lastmodified}: The datetime the files / folders was lastmodified.
    """

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="lastmodified",
        files=True,
        dirs=True,
    )

    def get_datetime(self, path: Path) -> datetime:
        return read_lastmodified(path)



from __future__ import annotations

from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render

from .common.conflict import ConflictMode, resolve_conflict


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Rename:

    """Renames a file.

    Attributes:
        name (str):
            The new name for the file / dir.

        on_conflict (str):
            What should happen in case **dest** already exists.
            One of `skip`, `overwrite`, `trash`, `rename_new` and `rename_existing`.
            Defaults to `rename_new`.

        rename_template (str):
            A template for renaming the file / dir in case of a conflict.
            Defaults to `{name} {counter}{extension}`.

    The next action will work with the renamed file / dir.
    """

    new_name: str
    on_conflict: ConflictMode = "rename_new"
    rename_template: str = "{name} {counter}{extension}"
    # TODO: keep_extension?

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="rename",
        standalone=False,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._new_name = Template.from_string(self.new_name)
        self._rename_template = Template.from_string(self.rename_template)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        assert res.path is not None, "Does not support standalone mode"
        new_name = render(self._new_name, res.dict())
        if "/" in new_name:
            raise ValueError(
                "The new name cannot contain slashes. "
                "To move files or folders use `move`."
            )
        dst = res.path.with_name(new_name)
        skip_action, dst = resolve_conflict(
            dst=dst,
            res=res,
            conflict_mode=self.on_conflict,
            rename_template=self._rename_template,
            simulate=simulate,
            output=output,
        )

        if skip_action:
            return

        output.msg(res=res, msg=f"Renaming to {new_name}", sender=self)
        if not simulate:
            res.path.rename(dst)
        res.path = dst
        res.walker_skip_pathes.add(dst)



import textwrap
from typing import ClassVar, Dict, Optional

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Python:

    """Execute python code.

    Attributes:
        code (str): The python code to execute.
        run_in_simulation (bool):
            Whether to execute this code in simulation mode (Default false).

    Variables of previous filters are available, but you have to use the normal python
    dictionary syntax `x = regex["my_group"]`.
    """

    code: str
    run_in_simulation: bool = False

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="python",
        standalone=True,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self.code = textwrap.dedent(self.code)

    def __usercode__(self, print, **kwargs) -> Optional[Dict]:
        raise NotImplementedError()

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        if simulate and not self.run_in_simulation:
            output.msg(
                res=res,
                msg="** Code not run in simulation. **",
                level="warn",
                sender=self,
            )
            return

        def _output_msg(*values, sep: str = " ", end: str = ""):
            """
            the print function for the use code needs to print via the current output
            """
            msg = f"{sep.join(str(x) for x in values)}{end}"
            output.msg(res=res, msg=msg, sender=self)

        # codegen the user function with arguments as available in the resource
        kwargs = ", ".join(res.dict().keys())
        func = f"def __userfunc(print, {kwargs}):\n"
        func += textwrap.indent(self.code, "    ")
        func += "\n\nself.__usercode__ = __userfunc"
        exec(func, globals().copy(), locals().copy())
        result = self.__usercode__(print=_output_msg, **res.dict())

        # deep merge the resulting dict
        if not (result is None or isinstance(result, dict)):
            raise ValueError("The python code must return None or a dict")

        if isinstance(result, dict):
            res.deep_merge(key=self.action_config.name, data=result)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional

from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm as RichConfirm
from rich.status import Status
from rich.theme import Theme

from organize.utils import ChangeDetector

from ._sender import sender_name
from .output import Level

if TYPE_CHECKING:
    from organize.resource import Resource

    from ._sender import SenderType


def _highlight_path(path: Path, base_style: str, main_style: str) -> str:
    base = str(f"{path.parent}/")
    main = str(path.name)
    return f"[{base_style}]{base}[/][{main_style}]{main}[/]"


def format_sender(sender: SenderType, standalone: bool) -> str:
    if not standalone:
        return f"    - ([pipeline.source]{sender_name(sender)}[/]) "
    return f"([pipeline.source]{sender_name(sender)}[/]) "


def format_info(msg: str) -> str:
    return f"[pipeline.msg]{msg}[/]"


def format_warn(msg: str) -> str:
    return f"[pipeline.warn]{msg}[/]"


def format_error(msg: str) -> str:
    return f"[pipeline.error]ERROR! {msg}[/]"


class Confirm(RichConfirm):
    @classmethod
    def set_error_msg(cls, msg: str) -> None:
        cls.validate_error_message = msg


class Default:
    def __init__(self, theme: Optional[Theme] = None):
        if theme is None:
            theme = Theme(
                {
                    "info": "dim cyan",
                    "warning": "yellow",
                    "error": "bold red",
                    "simulation": "bold green",
                    "status": "bold green",
                    "rule": "bold cyan",
                    "location.base": "green",
                    "location.main": "bold green",
                    "path.base": "dim green",
                    "path.main": "green",
                    "pipeline.source": "cyan",
                    "pipeline.msg": "",
                    "pipeline.warn": "yellow",
                    "pipeline.error": "bold red",
                    "pipeline.prompt": "bold yellow",
                    "summary.done": "bold green",
                    "summary.fail": "red",
                }
            )
        self.console = Console(theme=theme, highlight=False)

        self.status = Status("", console=self.console)
        self.det_rule = ChangeDetector()
        self.det_location = ChangeDetector()
        self.det_path = ChangeDetector()
        self.simulate = False

    def show_resource(self, res: Resource):
        # rule changed
        if self.det_rule.changed(res.rule):
            self.det_location.reset()
            self.det_path.reset()
            self.console.print()
            rule_name = f"Rule #{res.rule_nr}"
            if res.rule is not None and res.rule.name is not None:
                rule_name += f": {res.rule.name}"
            self.console.rule(f"[rule]:gear: {rule_name}", align="left", style="rule")

        # location changed
        if self.det_location.changed(res.basedir):
            self.det_path.reset()
            if res.basedir:
                path_str = _highlight_path(
                    Path(res.basedir),
                    "location.base",
                    "location.main",
                )
                self.console.print(path_str)

        # path changed
        if self.det_path.changed(res.path):
            relative_path = res.relative_path()
            if relative_path is not None:
                path_str = _highlight_path(relative_path, "path.base", "path.main")
                self.console.print(f"  {path_str}")

    def start(
        self,
        simulate: bool,
        config_path: Optional[Path],
        working_dir: Path,
    ) -> None:
        self.det_rule.reset()
        self.det_location.reset()
        self.det_path.reset()

        self.simulate = simulate
        if self.simulate:
            self.console.print(Panel("SIMULATION", style="simulation"))

        if working_dir.resolve() != Path(".").resolve():
            self.console.print(f'Working dir: "{working_dir}"')
        if config_path:
            self.console.print(f'Config: "{config_path}"')

        status_verb = "simulating" if simulate else "organizing"
        self.status.update(f"[status]{status_verb}[/]")
        self.status.start()

    def msg(
        self,
        res: Resource,
        msg: str,
        sender: SenderType,
        level: Level = "info",
    ) -> None:
        self.show_resource(res)
        msg_pre = format_sender(sender=sender, standalone=res.path is None)
        if level == "info":
            self.console.print(f"{msg_pre}{format_info(msg=msg)}")
        elif level == "error":
            self.console.print(f"{msg_pre}{format_error(msg=msg)}")
        elif level == "warn":
            self.console.print(f"{msg_pre}{format_warn(msg=msg)}")

    def confirm(
        self,
        res: Resource,
        msg: str,
        default: bool,
        sender: SenderType,
    ) -> bool:
        self.status.stop()
        self.show_resource(res)
        msg_pre = format_sender(sender=sender, standalone=res.path is None)
        error_msg = f"{msg_pre}[prompt.invalid]Please enter Y or N[/]"
        Confirm.set_error_msg(msg=error_msg)
        result = Confirm.ask(
            prompt=f"{msg_pre}[pipeline.prompt]{msg}[/]",
            console=self.console,
            default=default,
        )
        self.status.start()
        return result

    def end(self, success_count: int, error_count: int):
        self.status.stop()
        self.console.print()
        if success_count == 0 and error_count == 0:
            self.console.print("[summary.done]Nothing to do[/]")
        else:
            msg = (
                f"[summary.done]success {success_count}[/] / "
                f"[summary.fail]fail {error_count}[/]"
            )
            self.console.print(msg)
        if self.simulate:
            self.console.print(Panel("SIMULATION", style="simulation"))



__doc__ = """
organize - The file management automation tool.

Usage:
  organize run   [options] [<config>]
  organize sim   [options] [<config>]
  organize new   [<config>]
  organize edit  [<config>]
  organize check [<config>]
  organize debug [<config>]
  organize show  [--path|--reveal] [<config>]
  organize list
  organize docs
  organize --version
  organize --help

Commands:
  run        Organize your files.
  sim        Simulate organizing your files.
  new        Creates a default config.
  edit       Edit the config file with $EDITOR
  check      Check config file validity
  debug      Shows the raw config parsing steps.
  show       Print the config to stdout.
               Use --reveal to reveal the file in your file manager
               Use --path to show the path to the file
  list       Lists config files found in the default locations.
  docs       Open the documentation.

Options:
  <config>                        A config name or path to a config file
  -W --working-dir <dir>          The working directory
  -F --format (DEFAULT|JSONL)     The output format [Default: DEFAULT]
  -T --tags <tags>                Tags to run (eg. "initial,release")
  -S --skip-tags <tags>           Tags to skip
  -h --help                       Show this help page.
"""
import os
import sys
from functools import partial
from pathlib import Path
from typing import Literal, Optional, Set

from docopt import docopt
from pydantic import BaseModel, ConfigDict, Field, ValidationError, field_validator
from pydantic.functional_validators import BeforeValidator
from rich.console import Console
from rich.syntax import Syntax
from rich.table import Table
from typing_extensions import Annotated
from yaml.scanner import ScannerError

from organize import Config, ConfigError
from organize.find_config import ConfigNotFound, find_config, list_configs
from organize.output import JSONL, Default, Output

from .__version__ import __is_prerelease__, __version__

DOCS_RTD = "https://organize.readthedocs.io"
DOCS_GHPAGES = "https://tfeldmann.github.io/organize/"

EXAMPLE_CONFIG = f"""\
# organize configuration file
# {DOCS_RTD}

rules:
  - locations:
    filters:
    actions:
      - echo: "Hello, World!"
"""

Tags = Set[str]
OutputFormat = Annotated[
    Literal["default", "jsonl"], BeforeValidator(lambda v: v.lower())
]

console = Console()


def _open_uri(uri: str):
    import webbrowser

    webbrowser.open(uri)


def execute(
    config: Optional[str],
    working_dir: Optional[Path],
    format: OutputFormat,
    tags: Tags,
    skip_tags: Tags,
    simulate: bool,
) -> None:
    output = JSONL() if format == "jsonl" else Default()
    assert isinstance(output, Output)
    config_path = find_config(name_or_path=config)
    Config.from_path(config_path).execute(
        simulate=simulate,
        output=output,
        tags=tags,
        skip_tags=skip_tags,
        working_dir=working_dir or Path("."),
    )


def new(config: Optional[str]) -> None:
    try:
        config_path = find_config(config)
        console.print(
            f'Config "{config_path}" already exists.\n'
            r'Use "organize new \[name]" to create a config in the default location.'
        )
    except ConfigNotFound as e:
        assert e.init_path is not None
        e.init_path.parent.mkdir(parents=True, exist_ok=True)
        e.init_path.write_text(EXAMPLE_CONFIG, encoding="utf-8")
        console.print(f'Config "{e.init_path.stem}" created at "{e.init_path}"')


def edit(config: Optional[str]) -> None:
    config_path = find_config(config)
    editor = os.getenv("EDITOR")
    if editor:
        os.system(f'{editor} "{config_path}"')
    else:
        _open_uri(config_path.as_uri())


def check(config: Optional[str]) -> None:
    config_path = find_config(config)
    Config.from_path(config_path=config_path)
    console.print(f'No problems found in "{config_path}".')


def debug(config: Optional[str]) -> None:
    from rich.pretty import pprint

    config_path = find_config(config)
    pprint(
        Config.from_path(config_path=config_path),
        expand_all=True,
        indent_guides=False,
    )


def show(config: Optional[str], path: bool, reveal: bool) -> None:
    config_path = find_config(name_or_path=config)
    if path:
        print(config_path)
    elif reveal:
        _open_uri(config_path.parent.as_uri())
    else:
        syntax = Syntax(config_path.read_text(encoding="utf-8"), "yaml")
        console.print(syntax)


def list_() -> None:
    table = Table()
    table.add_column("Config")
    table.add_column("Path", no_wrap=True, style="dim")
    for path in list_configs():
        table.add_row(path.stem, str(path))
    console.print(table)


def docs() -> None:
    uri = DOCS_GHPAGES if __is_prerelease__ else DOCS_RTD
    print(f'Opening "{uri}"')
    _open_uri(uri=uri)


class CliArgs(BaseModel):
    model_config = ConfigDict(extra="forbid")

    # commands
    run: bool
    sim: bool
    new: bool
    edit: bool
    check: bool
    debug: bool
    show: bool
    list: bool
    docs: bool

    # run / sim options
    config: Optional[str] = Field(..., alias="<config>")
    working_dir: Optional[Path] = Field(..., alias="--working-dir")
    format: OutputFormat = Field("default", alias="--format")
    tags: Optional[str] = Field(..., alias="--tags")
    skip_tags: Optional[str] = Field(..., alias="--skip-tags")

    # show options
    path: bool = Field(False, alias="--path")
    reveal: bool = Field(False, alias="--reveal")

    # docopt options
    version: bool = Field(..., alias="--version")
    help: bool = Field(..., alias="--help")

    @field_validator("tags", "skip_tags", mode="after")
    @classmethod
    def split_tags(cls, val) -> Set[str]:
        if val is None:
            return set()
        return set(val.split(","))


def cli() -> None:
    arguments = docopt(
        __doc__,
        version=f"organize v{__version__}",
        default_help=True,
    )
    try:
        args = CliArgs.model_validate(arguments)
        _execute = partial(
            execute,
            config=args.config,
            working_dir=args.working_dir,
            format=args.format,
            tags=args.tags,
            skip_tags=args.skip_tags,
        )
        if args.run:
            _execute(simulate=False)
        elif args.sim:
            _execute(simulate=True)
        elif args.new:
            new(config=args.config)
        elif args.edit:
            edit(config=args.config)
        elif args.check:
            check(config=args.config)
        elif args.debug:
            debug(config=args.config)
        elif args.show:
            show(config=args.config, path=args.path, reveal=args.reveal)
        elif args.list:
            list_()
        elif args.docs:
            docs()
    except (ConfigError, ConfigNotFound) as e:
        console.print(f"[red]Error: Config problem[/]\n{e}")
        sys.exit(1)
    except ValidationError as e:
        console.print(f"[red]Error: Invalid CLI arguments[/]\n{e}")
        sys.exit(2)
    except ScannerError as e:
        console.print(f"[red]Error: YAML syntax error[/]\n{e}")
        sys.exit(3)


if __name__ == "__main__":
    cli()



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, Protocol, runtime_checkable

if TYPE_CHECKING:
    from pathlib import Path

    from organize.resource import Resource

    from ._sender import SenderType


Level = Literal["info", "warn", "error"]


@runtime_checkable
class Output(Protocol):
    """
    The protocol all of organize's outputs must adhere to.
    """

    def start(
        self,
        simulate: bool,
        config_path: Optional[Path],
        working_dir: Path,
    ) -> None:
        ...

    def msg(
        self,
        res: Resource,
        msg: str,
        sender: SenderType,
        level: Level = "info",
    ) -> None:
        ...

    def confirm(
        self,
        res: Resource,
        msg: str,
        default: bool,
        sender: SenderType,
    ) -> bool:
        ...

    def end(self, success_count: int, error_count: int) -> None:
        ...



from typing import Union

from organize.action import HasActionConfig
from organize.filter import HasFilterConfig

SenderType = Union[HasActionConfig, HasFilterConfig, str]


def sender_name(sender: SenderType) -> str:
    if isinstance(sender, HasFilterConfig):
        return sender.filter_config.name
    elif isinstance(sender, HasActionConfig):
        return sender.action_config.name
    return str(sender)



from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.action import ActionConfig
from organize.output import Output
from organize.resource import Resource
from organize.template import Template, render


@dataclass(config=ConfigDict(coerce_numbers_to_str=True, extra="forbid"))
class Confirm:

    """Ask for confirmation before continuing."""

    msg: str = "Continue?"
    default: bool = True

    action_config: ClassVar[ActionConfig] = ActionConfig(
        name="confirm",
        standalone=True,
        files=True,
        dirs=True,
    )

    def __post_init__(self):
        self._msg = Template.from_string(self.msg)

    def pipeline(self, res: Resource, output: Output, simulate: bool):
        msg = render(self._msg, res.dict())
        result = output.confirm(res=res, msg=msg, sender=self, default=self.default)
        if not result:
            raise StopIteration("Aborted")



import fnmatch
import os
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Union


@dataclass
class ReportSummary:
    success: int = 0
    errors: int = 0

    def __add__(self, other: "ReportSummary") -> "ReportSummary":
        return ReportSummary(
            success=self.success + other.success,
            errors=self.errors + other.errors,
        )


class ChangeDetector:
    def __init__(self):
        self._prev = None
        self._ready = False

    def changed(self, value: Any) -> bool:
        if not self._ready:
            self._prev = value
            self._ready = True
            return True
        else:
            changed = value != self._prev
            self._prev = value
            return changed

    def reset(self) -> None:
        self._ready = False


def expandvars(path: Union[str, Path]) -> Path:
    return Path(os.path.expandvars(path)).expanduser()


def glob_match(pattern: str, string: str, *, case_sensitive: bool = False) -> bool:
    if case_sensitive:
        return fnmatch.fnmatchcase(string, pattern)
    return fnmatch.fnmatch(string.lower(), pattern.lower())


def deep_merge(a: dict, b: dict, *, add_keys=True) -> dict:
    result = deepcopy(a)
    for bk, bv in b.items():
        av = result.get(bk)
        if isinstance(av, dict) and isinstance(bv, dict):
            result[bk] = deep_merge(av, bv, add_keys=add_keys)
        elif (av is not None) or add_keys:
            result[bk] = deepcopy(bv)
    return result


def deep_merge_inplace(base: dict, updates: dict) -> None:
    for bk, bv in updates.items():
        av = base.get(bk)
        if isinstance(av, dict) and isinstance(bv, dict):
            deep_merge_inplace(av, bv)
        else:
            base[bk] = bv



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, NamedTuple

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from .output import Output
    from .resource import Resource


class ActionConfig(NamedTuple):
    name: str
    standalone: bool
    files: bool
    dirs: bool


@runtime_checkable
class HasActionConfig(Protocol):
    action_config: ClassVar[ActionConfig]


class HasActionPipeline(Protocol):
    def pipeline(self, res: Resource, output: Output, simulate: bool):
        ...


@runtime_checkable
class Action(HasActionConfig, HasActionPipeline, Protocol):
    def __init__(self, *args, **kwargs) -> None:
        # allow any amount of args / kwargs for BaseModel and dataclasses.
        ...



from typing import ClassVar

from pydantic.config import ConfigDict
from pydantic.dataclasses import dataclass

from organize.filter import FilterConfig
from organize.output import Output
from organize.resource import Resource


@dataclass(config=ConfigDict(extra="forbid"))
class Empty:

    """Finds empty dirs and files"""

    filter_config: ClassVar[FilterConfig] = FilterConfig(
        name="empty",
        files=True,
        dirs=True,
    )

    def pipeline(self, res: Resource, output: Output) -> bool:
        return res.is_empty()
